"""
Testing script for volume.py
An ellipse is created and the analytical and numerical solution are compared
"""
import itertools as IT

import numpy as np

import dclab
from dclab.features.volume import get_volume

from helper_methods import retrieve_data


def area_of_polygon(x, y):
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def centroid_of_polygon(points):
    """
    http://stackoverflow.com/a/14115494/190597 (mgamba)
    Centroid of polygon:
    http://en.wikipedia.org/wiki/Centroid#Centroid_of_polygon
    """
    area = area_of_polygon(*zip(*points))
    result_x = 0
    result_y = 0
    N = len(points)
    points = IT.cycle(points)
    x1, y1 = next(points)
    for _i in range(N):
        x0, y0 = x1, y1
        x1, y1 = next(points)
        cross = (x0 * y1) - (x1 * y0)
        result_x += (x0 + x1) * cross
        result_y += (y0 + y1) * cross
    result_x /= (area * 6.0)
    result_y /= (area * 6.0)
    return abs(result_x), abs(result_y)


def get_ellipse_coords(a, b, x=0.0, y=0.0, angle=0.0, k=2):
    """ Draws an ellipse using (360*k + 1) discrete points; based on pseudo code
    given at http://en.wikipedia.org/wiki/Ellipse
    k = 1 means 361 points (degree by degree)
    a = major axis distance,
    b = minor axis distance,
    x = offset along the x-axis
    y = offset along the y-axis
    angle = clockwise rotation [in degrees] of the ellipse;
        * angle=0  : the ellipse is aligned with the positive x-axis
        * angle=30 : rotated 30 degrees clockwise from positive x-axis
    """
    pts = np.zeros((360 * k + 1, 2))

    beta = -angle * np.pi / 180
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)
    alpha = np.radians(np.r_[0.:360.:1j * (360 * k + 1)])

    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)

    pts[:, 0] = x + (a * cos_alpha * cos_beta - b * sin_alpha * sin_beta)
    pts[:, 1] = y + (a * cos_alpha * sin_beta + b * sin_alpha * cos_beta)

    return pts


def test_af_volume():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    vol = ds["volume"]
    # There are a lot of nans, because the contour is not given everywhere
    vol = vol[~np.isnan(vol)]
    assert np.allclose(vol[0], 574.60368907528346)
    assert np.allclose(vol[12], 1010.5669523203878)


def test_orientation():
    """counter-clockwise vs clockwise"""
    # Helper definitions to get an ellipse
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5,
                               angle=0,
                               k=100)
    # obtain the centroid (corresponds to pos_x and pos_lat)
    cx, cy = centroid_of_polygon(ellip)

    v1 = get_volume(cont=ellip,
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)

    # Turn contour around
    v2 = get_volume(cont=ellip[::-1, :],
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)

    assert np.all(v1 == v2)


def test_get_volume():
    # Helper definitions to get an ellipse
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5,
                               angle=0,
                               k=100)
    # obtain the centroid (corresponds to pos_x and pos_lat)
    cx, cy = centroid_of_polygon(ellip)
    elliplist = []
    elliplist.append(ellip)
    volume = get_volume(cont=elliplist,
                        pos_x=[cx],
                        pos_y=[cy],
                        pix=1)

    # Analytic solution for volume of ellipsoid:
    V = 4 / 3 * np.pi * major * minor**2
    msg = "Calculation of volume is faulty!"
    assert np.allclose(np.array(volume), np.array(V)), msg


def test_shape():
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5.0,
                               angle=0,
                               k=100)
    cx, cy = centroid_of_polygon(ellip)
    # no lists
    volume = get_volume(cont=ellip,
                        pos_x=cx,
                        pos_y=cy,
                        pix=1)
    assert isinstance(volume, float)

    volumelist = get_volume(cont=[ellip],
                            pos_x=[cx],
                            pos_y=[cy],
                            pix=1)
    assert isinstance(volumelist, np.ndarray)


def test_xpos():
    """xpos is not necessary to compute volume dense ellipse"""
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5.0,
                               angle=0,
                               k=100)
    cx, cy = centroid_of_polygon(ellip)
    # no lists
    v0 = get_volume(cont=ellip,
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)
    for cxi in np.linspace(0, 2 * cx, 10):
        vi = get_volume(cont=ellip,
                        pos_x=cxi,
                        pos_y=cy,
                        pix=1)
        assert np.allclose(v0, vi)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import abc
import collections
import uuid

import numpy as np

from .mllibs import tensorflow as tf
from ..rtdc_dataset.ancillaries import af_ml_score


class BaseModel(abc.ABC):
    def __init__(self, bare_model, inputs, outputs, model_name=None,
                 output_labels=None):
        """
        Parameters
        ----------
        bare_model:
            Underlying ML model
        inputs: list of str
            List of model input features, e.g.
            ``["deform", "area_um"]``
        outputs: list of str
            List of output features the model provides in that order, e.g.
            ``["ml_score_rbc", "ml_score_rt1", "ml_score_tfe"]``
        model_name: str or None
            The name of the models
        output_labels: list of str
            List of more descriptive labels for the features, e.g.
            ``["red blood cell", "type 1 cell", "troll cell"]``.
        """
        self.bare_model = bare_model
        self.inputs = inputs
        self.outputs = outputs
        self.name = model_name or str(uuid.uuid4())[:5]
        self.output_labels = output_labels or outputs

    def __enter__(self):
        self.register()
        return self

    def __exit__(self, *args):
        self.unregister()

    @staticmethod
    @abc.abstractmethod
    def supported_formats():
        """List of dictionaries containing model formats

        Returns
        -------
        fmts: list
            Each item contains the keys "name" (format name),
            "suffix" (saved file suffix), "requires" (Python
            dependencies).
        """

    @staticmethod
    @abc.abstractmethod
    def load_bare_model(path):
        """Load an implementation-specific model from a file

        This will set the `self.model` attribute. Make sure that
        the other attributes are set properly as well.
        """

    @staticmethod
    @abc.abstractmethod
    def save_bare_model(path, bare_model, save_format=None):
        """Save an implementation-specific model to a file

        Parameters
        ----------
        path: str or path-like
            Path to store model to
        bare_model: object
            The implementation-specific bare model
        save_format: str
            Must be in `supported_formats`
        """

    @abc.abstractmethod
    def predict(self, ds):
        """Return the probabilities of `self.outputs` for `ds`

        Parameters
        ----------
        ds: dclab.rtdc_dataset.RTDCBase
            Dataset to apply the model to

        Returns
        -------
        ofdict: dict
            Output feature dictionary with features as keys
            and 1d ndarrays as values.

        Notes
        -----
        This function calls :func:`BaseModel.get_dataset_features`
        to obtain the input feature matrix.
        """

    def get_dataset_features(self, ds, dtype=np.float32):
        """Return the dataset features used for inference

        Parameters
        ----------
        ds: dclab.rtdc_dataset.RTDCBase
            Dataset from which to retrieve the feature data
        dtype: dtype
            All features are cast to this dtype

        Returns
        -------
        fdata: 2d ndarray
            2D array of shape (len(ds), len(self.inputs));
            i.e. to access the array containing the first feature,
            for all events, you would do `fdata[:, 0]`.
        """
        fdata = np.zeros((len(ds), len(self.inputs)), dtype=dtype)
        for ii, feat in enumerate(self.inputs):
            fdata[:, ii] = ds[feat]
        return fdata

    def register(self):
        """Register this model to the dclab ancillary features"""
        af_ml_score.register(self)

    def unregister(self):
        """Unregister from dclab ancillary features"""
        af_ml_score.unregister(self)


class TensorflowModel(BaseModel):
    """Handle tensorflow models"""
    @staticmethod
    def supported_formats():
        return [{"name": "tensorflow-SavedModel",
                 "suffix": ".tf",
                 "requirements": "tensorflow"}
                ]

    @staticmethod
    def load_bare_model(path):
        """Load a tensorflow model"""
        # We don't use tf.saved_model.load, because it does not
        # return a keras layer.
        bare_model = tf.keras.models.load_model(str(path))
        return bare_model

    @staticmethod
    def save_bare_model(path, bare_model, save_format="tensorflow-SavedModel"):
        """Save a tensorflow model"""
        assert save_format == "tensorflow-SavedModel"
        tf.saved_model.save(obj=bare_model, export_dir=str(path))

    def predict(self, ds, batch_size=32):
        """Return the probabilities of `self.outputs` for `ds`

        Parameters
        ----------
        ds: dclab.rtdc_dataset.RTDCBase
            Dataset to apply the model to
        batch_size: int
            Batch size for inference with tensorflow

        Returns
        -------
        ofdict: dict
            Output feature dictionary with features as keys
            and 1d ndarrays as values.

        Notes
        -----
        Before prediction, this method asserts that the outputs of the
        model are converted to probabilities. If the final layer
        is one-dimensional and does not have a sigmoid activation,
        then a sigmoid activation layer is added (binary
        classification) ``tf.keras.layers.Activation("sigmoid")``.
        If the final layer has more dimensions and is not a
        ``tf.keras.layers.Softmax()`` layer, then a softmax layer
        is added.
        """
        probability_model = tf.keras.Sequential([self.bare_model])
        if self.bare_model.output_shape[1] > 1:
            # Multiple outputs; check for softmax
            if not self.has_softmax_layer():
                probability_model.add(tf.keras.layers.Softmax())
        else:
            # Binary classification; check for sigmoid
            if not self.has_sigmoid_activation():
                probability_model.add(tf.keras.layers.Activation("sigmoid"))

        fdata = self.get_dataset_features(ds)
        tfdata = tf.data.Dataset.from_tensor_slices(fdata).batch(batch_size)
        ret = probability_model.predict(tfdata)
        ofdict = collections.OrderedDict()
        for ii, key in enumerate(self.outputs):
            ofdict[key] = ret[:, ii]
        return ofdict

    def has_sigmoid_activation(self, layer_config=None):
        """Return True if final layer has "sigmoid" activation function"""
        if layer_config is None:
            layer_config = self.bare_model.get_config()
        if "layers" in layer_config:
            return self.has_sigmoid_activation(layer_config["layers"][-1])
        else:
            activation = layer_config.get("config", "").get("activation", "")
            return activation == "sigmoid"

    def has_softmax_layer(self, layer_config=None):
        """Return True if final layer is a Softmax layer"""
        if layer_config is None:
            layer_config = self.bare_model.get_config()
        if "layers" in layer_config:
            return self.has_softmax_layer(layer_config["layers"][-1])
        else:
            return layer_config["class_name"] == "Softmax"



from .ancillary_feature import AncillaryFeature


def register(dc_model):
    """Register an ML model

    For user convenience, this function can also be invoked by calling
    `dc_model.register`. Models can be removed from the registration
    by calling `unregister` or `dc_model.unregister`.

    Parameters
    ----------
    dc_model: dclab.ml.models.BaseModel
        ML model to register
    """
    # Check whether the model is already registered
    for af in list(AncillaryFeature.features):
        if af.data is dc_model:
            return
    # Check for feature existence
    for feat in dc_model.outputs:
        # Do not allow registration of the same feature at once
        for of in AncillaryFeature.features:
            if of.feature_name == feat:
                raise ValueError("Feature '{}' is already ".format(feat)
                                 + "defined by another ancillary feature "
                                 + "with the model '{}'.".format(of.data))

    # Register ml_score_??? features
    for feat in dc_model.outputs:
        # create an ancillary feature for each output
        def method(ds, feature=feat):
            return dc_model.predict(ds)[feature]
        AncillaryFeature(feature_name=feat,
                         method=method,
                         req_config=[],
                         req_features=dc_model.inputs,
                         data=dc_model,
                         )


def unregister(dc_model):
    """Unregister an ML model

    See :func:`register` for more information.
    """
    for af in list(AncillaryFeature.features):
        if af.data is dc_model:
            AncillaryFeature.features.remove(af)
            # Do not break here! There are multiple AncillaryFeatures
            # for one model.



from distutils.version import LooseVersion
import importlib

LIB_MIN_VERSIONS = {
    "tensorflow": "2.0",
}


class MockPackage(object):
    def __init__(self, name):
        self.name = name

    def __getattr__(self, item):
        raise ImportError("Please install '{}>={}'!".format(
            self.name, LIB_MIN_VERSIONS[self.name]))


for _lib in LIB_MIN_VERSIONS:
    try:
        _mod = importlib.import_module(_lib)
        _v_req = LIB_MIN_VERSIONS[_lib]
        if LooseVersion(_mod.__version__) < LooseVersion(_v_req):
            raise ValueError("Please install '{}>={}'!".format(_lib, _v_req))
    except ImportError:
        _mod = MockPackage(_lib)
    locals()[_lib] = _mod



from distutils.version import LooseVersion
import importlib

R_MIN_VERSION = "2.9.4"

R_SUBMODULES = [
    "rpy2.robjects.packages",
    "rpy2.situation",
    "rpy2.robjects.vectors",
]

R_SUBMODULES_3 = [
    "rpy2.rinterface_lib.callbacks",
]


class VersionError(BaseException):
    pass


class MockRPackage(object):
    def __getattr__(self, item):
        raise ImportError("Please install 'rpy2>={}'!".format(R_MIN_VERSION))


def import_r_submodules():
    importlib.import_module("rpy2.situation")
    if rpy2.situation.get_r_home() is None:
        return False
    else:
        if rpy2_is_version_3:
            mods = R_SUBMODULES + R_SUBMODULES_3
        else:
            mods = R_SUBMODULES
        for sm in mods:
            importlib.import_module(sm)


try:
    rpy2 = importlib.import_module("rpy2")
    if LooseVersion(rpy2.__version__) < LooseVersion(R_MIN_VERSION):
        raise VersionError("Please install 'rpy2>={}'!".format(R_MIN_VERSION))
except ImportError:
    rpy2 = MockRPackage()
    rpy2_is_version_3 = False
else:
    rpy2_is_version_3 = LooseVersion(rpy2.__version__) >= LooseVersion("3.0")
    import_r_submodules()



import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.features.bright import get_bright

from helper_methods import retrieve_data


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.event_image.'
                            + 'InitialFrameMissingWarning')
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_brightness():
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_avg = ds._events.pop("bright_avg")
    real_sd = ds._events.pop("bright_sd")
    # This will cause a zero-padding warning:
    comp_avg = ds["bright_avg"]
    comp_sd = ds["bright_sd"]
    idcompare = ~np.isnan(comp_avg)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_avg[idcompare], comp_avg[idcompare])
    assert np.allclose(real_sd[idcompare], comp_sd[idcompare])


def test_simple_bright():
    ds = new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    for ii in range(2, 7):
        # This stripped dataset has only 7 video frames / contours
        image = ds["image"][ii]
        mask = ds["mask"][ii]
        avg, std = get_bright(mask=mask, image=image, ret_data="avg,sd")
        assert np.allclose(avg, ds["bright_avg"][ii])
        assert np.allclose(std, ds["bright_sd"][ii])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""
This library contains classes and methods for the analysis
of real-time deformability cytometry (RT-DC) datasets.
"""
# flake8: noqa: F401
from . import definitions as dfn
from . import features
from . import isoelastics
from . import kde_contours
from . import kde_methods
from . import lme4
from . import ml
from .polygon_filter import PolygonFilter
from . import rtdc_dataset
from .rtdc_dataset import new_dataset
from .rtdc_dataset.feat_temp import register_temporary_feature, \
    set_temporary_feature
from . import statistics

from ._version import version as __version__



import numpy as np

import dclab

from helper_methods import example_data_dict, retrieve_data, \
    example_data_sets


def test_af_0basic():
    ds = dclab.new_dataset(retrieve_data(example_data_sets[1]))
    for cc in ['fl1_pos',
               'frame',
               'size_x',
               'size_y',
               'contour',
               'area_cvx',
               'circ',
               'image',
               'trace',
               'fl1_width',
               'nevents',
               'pos_x',
               'pos_y',
               'fl1_area',
               'fl1_max',
               ]:
        assert cc in ds

    # ancillaries
    for cc in ["deform",
               "area_um",
               "aspect",
               "frame",
               "index",
               "time",
               ]:
        assert cc in ds


def test_af_0error():
    keys = ["circ"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    try:
        ds["unknown_column"]
    except KeyError:
        pass
    else:
        raise ValueError("Should have raised KeyError!")


def test_af_aspect():
    # Aspect ratio of the data
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    aspect = ds["aspect"]
    assert np.sum(aspect > 1) == 904
    assert np.sum(aspect < 1) == 48


def test_af_area_ratio():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video.zip"))
    comp_ratio = ds["area_ratio"]
    # The convex area is always >= the raw area
    assert np.all(comp_ratio >= 1)
    assert np.allclose(comp_ratio[0], 1.0196464)


def test_af_deform():
    keys = ["circ"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert np.allclose(ds["deform"], 1 - ds["circ"])


def test_af_time():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    tt = ds["time"]
    assert tt[0] == 0
    assert np.allclose(tt[1], 0.0385)
    assert np.all(np.diff(tt) > 0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Reading and writing trained machine learning models for dclab"""
import atexit
import hashlib
import json
import pathlib
import shutil
import tempfile
import time
import traceback as tb
import warnings

from . import models

#: Supported file formats (including respective model classes).
SUPPORTED_FORMATS = {}
for _md in [models.TensorflowModel]:

    for _fmt in _md.supported_formats():
        SUPPORTED_FORMATS[_fmt["name"]] = {
            "requirements": _fmt["requirements"],
            "suffix": _fmt["suffix"],
            "class": _md}


class ModelFormatExportFailedWarning(UserWarning):
    pass


def export_model(path, model, enforce_formats=None):
    """Export an ML model to all possible formats

    The model must be exportable with at least one method
    listed in :const:`SUPPORTED_FORMATS`.

    Parameters
    ----------
    path: str or pathlib.Path
        Directory where the model is stored to. For each supported
        model, a new subdirectory or file is created.
    model: An instance of an ML model, NOT dclab.ml.models.BaseModel
        Trained model instance
    enforce_formats: list of str
        Enforced file formats for export. If the export for one
        of these file formats fails, a ValueError is raised.
    """
    if enforce_formats is None:
        enforce_formats = []
    path = pathlib.Path(path)
    if not path.is_dir():
        raise ValueError(
            "Output `path` should be a directory: '{}'".format(path))
    if len(sorted(path.rglob("*"))) != 0:
        raise ValueError(
            "Model output directory should be empty: '{}'".format(path))
    if enforce_formats:
        for e_fmt in enforce_formats:
            if e_fmt not in SUPPORTED_FORMATS:
                raise ValueError(
                    "Unsupported format '{}', expected ".format(e_fmt)
                    + "one of {}!".format(", ".join(SUPPORTED_FORMATS.keys())))
    exported_formats = {}
    for fmt in SUPPORTED_FORMATS:
        tmp = tempfile.mkdtemp(prefix="dclab_ml_{}".format(fmt))
        try:
            suffix = SUPPORTED_FORMATS[fmt]["suffix"]
            tmp_out = pathlib.Path(tmp) / (fmt + suffix)
            cls = SUPPORTED_FORMATS[fmt]["class"]
            cls.save_bare_model(tmp_out, model, save_format=fmt)
            # attempt to load the model to see if it worked
            cls.load_bare_model(tmp_out)
        except BaseException:
            warnings.warn("Could not export to '{}': {}".format(
                fmt, tb.format_exc(limit=1)),
                ModelFormatExportFailedWarning)
            if fmt in enforce_formats:
                raise
        else:
            pout = path / tmp_out.name
            pathlib.Path(tmp_out).rename(pout)
            exported_formats[fmt] = tmp_out.name
        finally:
            shutil.rmtree(tmp, ignore_errors=True)

    if not exported_formats:
        raise ValueError("Export failed for all model formats!")

    # Now compute the hash and get the destination filename
    m_hash = hash_path(path)
    m_dict = {"sha256": m_hash,
              "formats": exported_formats,
              "date": time.strftime("%Y-%m-%d %H:%M"),
              }
    return m_dict


def hash_path(path):
    """Create a SHA256 hash of a file or all files in a directory

    The files are sorted before hashing for reproducibility.
    """
    path = pathlib.Path(path)
    assert path.is_dir()
    hasher = hashlib.sha256()
    if path.is_dir():
        paths = sorted(path.rglob("*"))
    else:
        paths = [path]
    for pp in paths:
        hasher.update(pp.name.encode())
        if pp.is_dir():
            continue
        with pp.open("rb") as fd:
            while True:
                chunk = fd.read(65536)
                if chunk:
                    hasher.update(chunk)
                else:
                    break
    return hasher.hexdigest()


def load_modc(path, from_format=None):
    """Load models from a .modc file for inference

    Parameters
    ----------
    path: str or path-like
        Path to a .modc file
    from_format: str
        If set to None, the first available format in
        :const:`SUPPORTED_FORMATS` is used. If set to
        a key in :const:`SUPPORTED_FORMATS`, then this
        format will take precedence and an error will
        be raised if loading with this format fails.

    Returns
    -------
    model: dclab.ml.models.BaseModel
        Model that can be used for inference via `model.predict`
    """
    # unpack everything
    t_dir = pathlib.Path(tempfile.mkdtemp(prefix="modc_load_"))
    cleanup = atexit.register(lambda: shutil.rmtree(t_dir, ignore_errors=True))
    shutil.unpack_archive(path, t_dir, format="zip")

    # Get the metadata
    with (t_dir / "index.json").open("r") as fd:
        meta = json.load(fd)

    assert meta["model count"] == len(meta["models"])
    for model_dict in meta["models"]:
        mpath = t_dir / model_dict["path"]

        formats = list(model_dict["formats"].keys())
        if from_format:
            formats = [from_format] + formats

        for fmt in formats:
            if fmt in SUPPORTED_FORMATS:
                cls = SUPPORTED_FORMATS[fmt]["class"]
                load = cls.load_bare_model
                try:
                    bare_model = load(mpath / model_dict["formats"][fmt])
                except BaseException:
                    if from_format and fmt == from_format:
                        # user requested this format explicitly
                        raise
                else:
                    # load `bare_model` into BaseModel
                    model = cls(bare_model=bare_model,
                                inputs=model_dict["input features"],
                                outputs=model_dict["output features"],
                                output_labels=model_dict["output labels"],
                                model_name=model_dict["name"]
                                )
                    break
            elif from_format and fmt == from_format:
                raise ValueError("The format specified via `from_format` "
                                 + " '{}' is not supported!".format(fmt))
        else:
            raise ValueError("No compatible model file format found!")

    # We are nice and do the cleanup before exit
    cleanup()
    atexit.unregister(cleanup)

    return model


def save_modc(path, dc_models):
    """Save ML models to a .modc file

    Parameters
    ----------
    path: str, pathlib.Path
        Output .modc path
    dc_models: list of dclab.ml.models.BaseModel or dclab.ml.models.BaseModel
        Models to save

    Returns
    -------
    meta: dict
        Dictionary written to index.json in the .modc file
    """
    if not isinstance(dc_models, list):
        dc_models = [dc_models]

    # save all models to a temporary directory
    t_dir = pathlib.Path(tempfile.mkdtemp(prefix="modc_save_"))
    cleanup = atexit.register(lambda: shutil.rmtree(t_dir, ignore_errors=True))

    model_data = []
    for ii, mm in enumerate(dc_models):
        p_mod = t_dir / "model_{}".format(ii)
        p_mod.mkdir()
        m_dict = export_model(p_mod, mm.bare_model)
        m_dict["index"] = ii
        m_dict["input features"] = mm.inputs
        m_dict["output features"] = mm.outputs
        m_dict["output labels"] = mm.output_labels
        m_dict["name"] = mm.name
        m_dict["path"] = p_mod.name
        model_data.append(m_dict)

    meta = {
        "model count": len(dc_models),
        "models": model_data,
    }

    # save metadata
    with (t_dir / "index.json").open("w") as fd:
        json.dump(meta, fp=fd, indent=2, separators=(',', ': '),
                  sort_keys=True)

    aout = shutil.make_archive(base_name=path, format="zip", root_dir=t_dir,
                               base_dir=".")
    pathlib.Path(aout).rename(path)

    # We are nice and do the cleanup before exit
    cleanup()
    atexit.unregister(cleanup)

    return meta
